//=====================================
// GameBoy Advance Keyboard
//
//  by Jeff Frohwein
// http://www.devrs.com
//
//  This software uses the following hardware interface
//     gbair.gif
//  in order to use InfraRed keyboard input from
//  the Sony Wireless Keyboard Model# SWK-8640 that
//  is used with WebTV. This keyboard can be ordered from
// http://www.mcmelectronics.com as part# KIW200 for ~US$46
// or from many other places for probably much cheaper.
// Probably any WebTV compatible IR keyboard will work even
// though this is the only version that was tested.

//   NOTE: THIS CIRCUIT IS VERY SENSITIVE
//  TO EXTERNAL LIGHT. TRY IT WITHOUT A DIRECT LIGHT SOURCE
//  AROUND. A better design would use an IR module (with a
//  carrier frequency filter) instead of a simple IR transistor.
//
//  Compiled with GCC 2.9x.
// The "crtls" package from devrs.com/gba may be required
// to get the proper lnkscript & crt0.s to compile this file.
//
//  No state handling is done in this package. (i.e. Pressing
// shift key processing must be done in your own code to convert
// lower case to upper case.) Shift, Ctrl, Alt, Fn and other
// buttons return raw values, currently, rather than useful
// values. Interrupts can not be used with the current setup
// unless you disable interrupts for 22*0.8 milliseconds since
// it would interfere with the received serial data. You must
// devise your own timer interrupt algorithm if you want to
// allow other interrupts to be used.
//
//
//  The WebTV bit stream consists of a start bit, 20 bits of
// data, and a stop bit. Each bit is 0.8 milliseconds wide.
// All other know info is contained in this file. YOU must
// do your own decoding to learn/gather more info about this
// keyboard.
//
//=====================================

// v1.0 - Original version

#include "main.h"
#include "ibm.h"

MULTIBOOT

int CursorX,CursorY;

void gfxPoint (int x, int y, int clr, int scrnaddr)
   {
   *(u16 *)((x<<1) + (y*480) + scrnaddr) = clr;
   }

void putc2 (int c)
   {
   int row,x,y;

   for (y=0; y<8; y++)
      {
      row = IBMFont[(c<<3) + y];
      for (x=0; x<8; x++)
         {
         if (row & 0x80)
            gfxPoint((CursorX<<3)+x, (CursorY<<3)+y, 0x0000, VRAM);
         else
            gfxPoint((CursorX<<3)+x, (CursorY<<3)+y, 0x7fff, VRAM);
         row = (row << 1) & 0xff;
         }
      }
   CursorX++;
   if (CursorX>29)
      {
      CursorX = 0;
      CursorY++;
      }
   }

void CrLf (void)
   {
   CursorX = 0;
   CursorY++;
   }

void Hex1 (u32 v)
   {
   if (v < 10)
      putc2(v+0x30);
   else
      putc2(v-10+0x61);
   }

void Hex2 (u32 v)
   {
   Hex1((v>>4) & 15);
   Hex1(v & 15);
   }

void Hex7 (u32 v)
   {
   Hex1((v>>24) & 0xff);
   Hex2((v>>16) & 0xff);
   Hex2((v>>8) & 0xff);
   Hex2(v & 0xff);
   }

// The keyboard read routine is put in Internal WRAM
// since IWRAM always runs at 0 wait states. It could
// also reside in ROM but the timing would change
// (break) depending on the ROM speed setting unless
// some form of timing compensation is added.

int KeyboardRead (int *FuncAddr) IN_IWRAM;

int KeyboardRead (int *FuncAddr)
   {
   asm("
      push {r4}

   @ Set SD input mode
      ldr  r1,=0x4000134
      ldr  r2,=0x8000
      strh r2,[r1]
      mov  r3,#2

   @ Wait until SD goes low
   KRLoop:
      ldrh r2,[r1]
      and  r2,r3
      bne  KRLoop

   @ delay 6711 cycles to get to middle of bit
      ldr  r2,=1677
   KRHBLoop:
      sub  r2,#1     @ 1
      bne  KRHBLoop  @ 3

      mov r0,#0
      mov r3,#0
      mov r4,#21

   KRBitLoop:

   @ get bit
      lsl  r0,r0,#1  @ 2

      ldrh r2,[r1]   @ 3
      lsr  r2,r2,#2  @ 2
      adc  r0,r3     @ 1

   @ delay 13422-12 cycles to get to next bit
      ldr  r2,=3352
   KRFBLoop:
      sub  r2,#1       @ 1
      bne  KRFBLoop    @ 3

      sub  r4,#1     @ 1
      bne  KRBitLoop @ 3

      pop  {r4}
      bx   lr
   ");
   return (0);
   }

const char RawKeys2Ascii [] = {
  '?','?','?','?', '?','?','?','?', 'b','?','?','?', '?','?','?','?', //000-00f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?','?','?','?', //010-01f
  '?','?','?','?', '?',8  ,'?','?', '?','?','?','?', '/','?','?','?', //020-02f
  '?','?','?','?', '?','?','?','?', ' ','?','?','?', '?','n','?','?', //030-03f
  '~','?','?','?', '?','?','?','?', '?','5','?','?', '?','?','?','?', //040-04f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?','=','?','?', //050-05f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?','_','?','?', //060-06f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '6','?','?','?', //070-07f
  '?','?','?','?', '?','?','?','?', '?','v','?','?', '.','?','?','?', //080-08f
  '?','c','?','?', '?','?','?','?', '?','?','?','?', '?',',','?','?', //090-09f
  '?','x','?','?', '?','?','?','?', '?','?','?','?', '?','?','?','?', //0a0-0af
  'z','?','?','?', '?','?','?','?', '?',13 ,'?','?', 'm','?','?','?', //0b0-0bf
  '?','?','?','?', '?','?','?','?', 'f','?','?','?', '?','l','?','?', //0c0-0cf
  'd','?','?','?', '?','?','?','?', '?','?','?','?', 'k','?','?','?', //0d0-0df
  's','?','?','?', '?','?','?','?', '?','?','?','?', ';','?','?','?', //0e0-0ef
  '?','a','?','?', '?','?','?','?', 92 ,'?','?','?', '?','j','?','"', //0f0-0ff
  '?','?','?','?', '?','?','?','?', '?','t','?','?', '?','?','?','?', //100-10f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?',']','?','?', //110-11f
  '?','?','?','?', '?','?','?','?', 27 ,'?','?','?', '?','[','?','?', //120-12f
  9  ,'?','?','?', '?','?','?','?', '?',127,'?','?', 'y','?','?','?', //130-13f
  '?','?','?','?', '?','?','?','?', '4','?','?','?', '?','9','?','?', //140-14f
  '3','?','?','?', '?','?','?','?', '?','?','?','?', '8','?','?','?', //150-15f
  '2','?','?','?', '?','?','?','?', '?','?','?','?', '0','?','?','?', //160-16f
  '?','1','?','?', '?','?','?','?', '?','?','?','?', '?','7','?','?', //170-17f
  '?','?','?','?', '?','?','?','?', 'g','?','?','?', '?','?','?','?', //180-18f
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?','?','?','?', //190-19f
  '?','?','?','?', '?','?','?','?', '?','?','?','?','\'','?','?','?', //1a0-1af
  '?','?','?','?', '?','?','?','?', '?','?','?','?', '?','h','?','?', //1b0-1bf
  '?','?','?','?', '?','?','?','?', '?','r','?','?', 'o','?','?','?', //1c0-1cf
  '?','e','?','?', '?','?','?','?', '?','?','?','?', '?','i','?','?', //1d0-1df
  '?','w','?','?', '?','?','?','?', '?','?','?','?', '?','p','?','?', //1e0-1ef
  'q','?','?','?', '?','?','?','?', '?','?','?','?', 'u','?','?','?'  //1f0-1ff
  };

int AgbMain(void)
   {
   int key;
   int i;
   u16 keyb;

   CursorX = 0;
   CursorY = 0;
//   *(vu16 *)REG_WAITCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT; // 3-1 Wait Access Setting

   // Setup mode 3 graphics
   *(u16 *)REG_DISPCNT = DISP_MODE_3 | DISP_BG2_ON;
   *(u16 *)REG_STAT = 0;

   // Clear VRAM
   for (i=0; i<240*160*2; i=i+2)
      *(u16 *)(VRAM + i) = 0;

   while (1)
      {
      key = __FarFunction (KeyboardRead); // & 0x1ff;
      keyb = key & 0x1ff;

      // Uncomment the following line to display keyboard raw value
//      Hex7 (key & 0x1ff); CrLf();

      if ( ((key & 0xfffffe00) == 0x29400) &&
           (keyb != 0x98) &&
           (keyb != 0x118) )
        {
        // Display pressed key pressed
        putc2 (RawKeys2Ascii[key & 0x1ff]);
        }

      }

   }

